import string
import logging
import pathlib
from .smb import *
from .file import *
from .util import *
from .errors import *
import multiprocessing
from shutil import move
from .processpool import *
from traceback import format_exc


log = logging.getLogger('manspider.spiderling')


class SpiderlingMessage:
    '''
    Message which gets sent back to the parent through parent_queue
    '''

    def __init__(self, message_type, target, content):
        '''
        "message_type" is a string, and can be:
            "e" - error
            "a" - authentication failure
        '''
        self.type = message_type
        self.target = target
        self.content = content



class Spiderling:
    '''
    Enumerates SMB shares and spiders all possible directories/filenames up to maxdepth
    Designed to be threadable
    '''

    # these extensions don't get parsed for content, unless explicitly specified
    dont_parse = [
        '.png',
        '.gif',
        '.tiff',
        '.msi',
        '.bmp',
        '.jpg',
        '.jpeg',
        '.zip',
        '.gz',
        '.bz2',
        '.7z',
        '.xz',
    ]

    def __init__(self, target, parent):

        try:

            self.parent = parent
            self.target = target

            # unless we're only searching local files, connect to target
            if type(self.target) == pathlib.PosixPath:
                self.local = True
                self.go()

            else:
                self.local = False

                self.smb_client = SMBClient(
                    target,
                    parent.username,
                    parent.password,
                    parent.domain,
                    parent.nthash,
                    parent.use_kerberos,
                    parent.aes_key,
                    parent.dc_ip
                )

                logon_result = self.smb_client.login()
                if logon_result not in [True, None]:
                    self.message_parent('a', logon_result)

                if logon_result is not None:
                    self.go()

            # file parsing parallelized one process at a time
            # allows file to be parsed while next one is being fetched
            self.parser_process = None

        except KeyboardInterrupt:
            log.critical('Spiderling Interrupted')

        # log all exceptions
        except Exception as e:
            if log.level <= logging.DEBUG:
                log.error(format_exc())
            else:
                log.error(f'Error in spiderling: {e}')


    def go(self):
        '''
        go spider go spider go
        '''

        # local files
        if self.local:
            if self.parent.parser.content_filters:
                self.parse_local_files(self.files)
            else:
                # just list the files
                list(self.files)

        else:
            # remote files
            for file in self.files:

                # if content searching is enabled, parse the file
                if self.parent.parser.content_filters:
                    try:
                        self.parser_process.join()
                    except AttributeError:
                        pass
                    self.parser_process = multiprocessing.Process(target=self.parse_file, args=(file,))
                    self.parser_process.start()

                # otherwise, just save it
                elif not self.local:
                    log.info(f'{self.target}: {file.share}\\{file.name} ({bytes_to_human(file.size)})')
                    if not self.parent.no_download:
                        self.save_file(file)

        log.info(f'Finished spidering {self.target}')



    @property
    def files(self):
        '''
        Yields all files on the target to be parsed/downloaded
        Premptively download matching files into temp directory
        '''

        if self.local:
            for file in list(list_files(self.target)):
                if self.extension_blacklisted(file):
                    log.debug(f'{self.target}: Skipping {file}: extension is blacklisted')
                    continue
                if self.path_match(file) or (self.parent.or_logic and self.parent.parser.content_filters):
                    if self.path_match(file):
                        log.debug(pathlib.Path(file).relative_to(self.target))
                    if not self.is_binary_file(file):
                        yield file
                else:
                    log.debug(f'Skipping {file}: does not match filename/extension filters')

        else:
            for share in self.shares:
                for remote_file in self.list_files(share):
                    if not self.parent.no_download or self.parent.parser.content_filters:
                        self.get_file(remote_file)
                    yield remote_file



    def parse_file(self, file):
        '''
        Simple wrapper around self.parent.parser.parse_file()
        For sole purpose of threading
        '''

        try:

            if type(file) == RemoteFile:
                matches = self.parent.parser.parse_file(str(file.tmp_filename), pretty_filename=str(file))
                if matches and not self.parent.no_download:
                    self.save_file(file)
                else:
                    file.tmp_filename.unlink()

            else:
                log.debug(f'Found file: {file}')
                self.parent.parser.parse_file(file, file)

        # log all exceptions
        except Exception as e:
            if log.level <= logging.DEBUG:
                log.error(format_exc())
            else:
                log.error(f'Error parsing file {file}: {e}')

        except KeyboardInterrupt:
            log.critical('File parsing interrupted')


    @property
    def shares(self):
        '''
        Lists all shares on single target
        '''

        for share in self.smb_client.shares:
            if self.share_match(share):
                yield share



    def list_files(self, share, path='', depth=0, tries=2):
        '''
        List files inside a specific directory
        Only yield files which conform to all filters (except content)
        '''

        if depth < self.parent.maxdepth and self.dir_match(path):

            files = []
            while tries > 0:
                try:
                    files = list(self.smb_client.ls(share, path))
                    break
                except FileListError as e:
                    if 'ACCESS_DENIED' in str(e):
                        log.debug(f'{self.target}: Error listing files: {e}')
                        break
                    else:
                        tries -= 1

            if files:
                log.debug(f'{self.target}: {share}{path}: contains {len(files):,} items')

            for f in files:
                name = f.get_longname()
                full_path = f'{path}\\{name}'
                # if it's a directory, go deeper
                if f.is_directory():
                    for file in self.list_files(share, full_path, (depth+1)):
                        yield file

                else:

                    # skip the file if it didn't match extension filters
                    if self.extension_blacklisted(name):
                        log.debug(f'{self.target}: Skipping {share}{full_path}: extension is blacklisted')
                        continue

                    if not self.path_match(name):
                        if not (
                                # all of these have to be true in order to get past this point
                                # "or logic" is enabled
                                self.parent.or_logic and
                                # and file does not have a "don't parse" extension
                                (not self.is_binary_file(name)) and
                                # and content filters are enabled
                                self.parent.parser.content_filters
                            ):
                            log.debug(f'{self.target}: Skipping {share}{full_path}: filename/extensions do not match')
                            continue

                    # try to get the size of the file
                    try:
                        filesize = f.get_filesize()
                    except Exception as e:
                        self.smb_client.handle_impacket_error(e)
                        continue

                    # make the RemoteFile object (the file won't be read yet)
                    full_path_fixed = full_path.lstrip('\\')
                    remote_file = RemoteFile(full_path_fixed, share, self.target, size=filesize)

                    # if it's a non-empty file that's smaller than the size limit
                    if filesize > 0 and filesize < self.parent.max_filesize:

                        # if it matched filename/extension filters and we're downloading files
                        if (self.parent.file_extensions or self.parent.filename_filters) and not self.parent.no_download:
                            # but the extension is marked as "don't parse"
                            if self.is_binary_file(name):
                                # don't parse it, instead save it and continue
                                log.info(f'{self.target}: {remote_file.share}\\{remote_file.name}')
                                if self.get_file(remote_file):
                                    self.save_file(remote_file)
                                    continue

                        # file is ready to be parsed
                        yield remote_file

                    else:
                        log.debug(f'{self.target}: {full_path} is either empty or too large')


    def path_match(self, file):
        '''
        Based on whether "or" logic is enabled, return True or False
        if the filename + extension meets the requirements
        '''
        filename_match = self.filename_match(file)
        extension_match = self.extension_whitelisted(file)
        if self.parent.or_logic:
            return (filename_match and self.parent.filename_filters) or (extension_match and self.parent.file_extensions)
        else:
            return filename_match and extension_match



    def share_match(self, share):
        '''
        Return true if "share" matches any of the share filters
        '''

        # if the share has been whitelisted
        if ((not self.parent.share_whitelist) or (share.lower() in self.parent.share_whitelist)):
            # and hasn't been blacklisted
            if ((not self.parent.share_blacklist) or (share.lower() not in self.parent.share_blacklist)):
                return True
            else:
                log.debug(f'{self.target}: Skipping blacklisted share: {share}')
        else:
            log.debug(f'{self.target}: Skipping share {share}: not in whitelist')

        return False


    def dir_match(self, path):
        '''
        Return true if "path" matches any of the directory filters
        '''

        # convert forward slashes to backwards
        dirname = str(path).lower().replace('/', '\\')

        # root path always passes
        if not path:
            return True

        # if whitelist check passes
        if (not self.parent.dir_whitelist) or any([k.lower() in dirname for k in self.parent.dir_whitelist]):
            # and blacklist check passes
            if (not self.parent.dir_blacklist) or not any([k.lower() in dirname for k in self.parent.dir_blacklist]):
                return True
            else:
                log.debug(f'{self.target}: Skipping blacklisted dir: {path}')
        else:
            log.debug(f'{self.target}: Skipping dir {path}: not in whitelist')

        return False


    def filename_match(self, filename):
        '''
        Return true if "filename" matches any of the filename filters
        '''

        if (not self.parent.filename_filters) or any([f_regex.match(str(pathlib.Path(filename).stem)) for f_regex in self.parent.filename_filters]):
            return True
        else:
            log.debug(f'{self.target}: {filename} does not match filename filters')

        return False


    def is_binary_file(self, filename):
        '''
        Returns true if file is a bad extension type, e.g. encrypted or compressed
        '''

        extension = ''.join(pathlib.Path(filename).suffixes).lower()
        if any([extension.endswith(e.lower()) for e in self.dont_parse]):
            if extension not in self.parent.file_extensions:
                log.debug(f'{self.target}: Not parsing {filename} due to undesirable extension')
                return True
        return False


    def extension_blacklisted(self, filename):
        '''
        Return True if folder, file name, or extension has been blacklisted
        '''
        extension = ''.join(pathlib.Path(filename).suffixes).lower()
        excluded_extensions = list(self.parent.extension_blacklist)

        if not excluded_extensions:
            return False

        if not any([extension.endswith(e) for e in excluded_extensions]):
            return False
        else:
            log.debug(f'{self.target}: Skipping file with blacklisted extension: {filename}')
            return True


    def extension_whitelisted(self, filename):
        '''
        Return True if file extension has been whitelisted
        '''
        # a .tar.gz file will match both filters ".gz" and ".tar.gz"
        extension = ''.join(pathlib.Path(filename).suffixes).lower()
        extensions = list(self.parent.file_extensions)

        if not extensions:
            return True

        # if whitelist check passes
        if  any([(extension.endswith(e) if e else extension == e) for e in extensions]):
            log.debug(f'{self.target}: {filename} matches extension filters')
            return True
        else:
            log.debug(f'{self.target}: Skipping file {filename}, does not match extension filters')
            return False


    def message_parent(self, message_type, content=''):
        '''
        Send a message to the parent spider
        '''

        self.parent.spiderling_queue.put(
            SpiderlingMessage(message_type, self.target, content)
        )


    def parse_local_files(self, files):

        with ProcessPool(self.parent.threads) as pool:
            for r in pool.map(self.parse_file, files):
                pass


    def save_file(self, remote_file):
        '''
        Moves a file from temp storage into the loot directory
        '''

        allowed_chars = string.ascii_lowercase + string.ascii_uppercase + string.digits + '._ '

        # replace backslashes with underscores to preserve directory names
        loot_filename = str(remote_file).replace('\\', '_')
        # remove weird characters
        loot_filename = ''.join([c for c in loot_filename if c in allowed_chars])
        loot_dest = self.parent.loot_dir / loot_filename
        try:
            move(str(remote_file.tmp_filename), str(loot_dest))
        except Exception:
            log.warning(f'Error saving {remote_file}')


    def get_file(self, remote_file):
        '''
        Attempts to retrieve "remote_file" from share and returns True if successful
        '''

        try:
            smb_client = self.parent.get_smb_client(self.target)
            log.debug(f'{self.target}: Downloading {remote_file.share}\\{remote_file.name}')
            remote_file.get(smb_client)
            return True
        except FileRetrievalError as e:
            log.debug(f'{self.target}: {e}')

        return False




import logging
import traceback
from time import sleep
import multiprocessing as mp
from queue import Empty, Full


# set up logging
log = logging.getLogger('manspider.processpool')



class ProcessPool:
    '''
    usage:
    with ProcessPool(2) as pool:
        for i in pool.map(target, iterable):
            yield i
    '''

    def __init__(self, processes=None, daemon=False, name=''):

        self.name = f'ProcessPool'
        if name:
            self.name += f'-{name}'

        if processes is None:
            processes = mp.cpu_count()

        self.processes = processes
        self.pool = [None] * self.processes

        self.started_counter = 0
        self.finished_counter = 0

        self.daemon = daemon

        # make the result queue
        self.result_queue = mp.Manager().Queue()


    def map(self, func, iterable, args=(), kwargs={}):

        # loop until we're out of work
        for entry in iterable:

            try:

                while 1:

                    for result in self.results:
                        yield result

                    # start processes
                    for i in range(len(self.pool)):
                        process = self.pool[i]
                        if process is None or not process.is_alive():
                            self.pool[i] = mp.Process(target=self.execute, args=(func, self.result_queue, (entry,)+args), \
                                kwargs=kwargs, daemon=self.daemon)
                            self.pool[i].start()
                            self.started_counter += 1
                            log.debug(f'{self.name}: {self.started_counter:,} processes started')
                            # success, move on to next
                            assert False

                        for result in self.results:
                            yield result

                    # prevent unnecessary CPU usage
                    sleep(.1)

            except AssertionError:
                continue

        # wait for processes to finish
        while 1:

            finished_threads = [p is None or not p.is_alive() for p in self.pool]
            if all(finished_threads):
                self.finished_counter += len([p for p in self.pool if p is not None and not p.is_alive()])
                break
            else:
                log.debug(f'{self.name}: Waiting for {finished_threads.count(False):,} threads to finish')
                sleep(1)

        for result in self.results:
            yield result


    @property
    def results(self):

        while 1:
            try:
                result = self.result_queue.get_nowait()
                self.finished_counter += 1
                log.debug(f'{self.name}: {self.finished_counter:,} processes finished')
                yield result
            except Empty:
                sleep(.1)
                break


    @staticmethod
    def execute(func, result_queue, args=(), kwargs={}):
        '''
        Executes given function and places return value in result queue
        '''

        try:
            result_queue.put(func(*args, **kwargs))
        except Exception as e:
            if type(e) not in [FileNotFoundError]:
                log.critical(format_exc())
        except KeyboardInterrupt as e:
            log.critical('ProcessPool Interrupted')


    @staticmethod
    def _close_queue(q):

        while 1:
            try:
                q.get_nowait()
            except Empty:
                break
        q.close()


    def __enter__(self):

        return self


    def __exit__(self, exception_type, exception_value, traceback):

        try:
            self._close_queue(self.result_queue)
        except Exception:
            pass


import os
import magic
import string
import random
import logging
import ipaddress
from pathlib import Path

log = logging.getLogger('manspider.util')


def str_to_list(s):

    l = set()
    # try to open as file
    try:
        with open(s) as f:
            lines = set([l.strip() for l in f.readlines()])
            for line in lines:
                if line:
                    l.add(line)
    except OSError:
        l.add(s)

    return list(l)


def make_targets(s):
    '''
    Accepts filename, CIDR, IP, hostname, file, or folder
    Returns list of targets as IPs, hostnames, or Path() objects
    '''

    targets = set()

    p = Path(s)
    if s.lower() == 'loot':
        targets.add(Path.home() / '.manspider' / 'loot')

    elif p.is_dir():
        targets.add(p)

    else:
        for i in str_to_list(s):
            try:
                for ip in ipaddress.ip_network(i, strict=False):
                    targets.add(str(ip))
            except ValueError:
                targets.add(i)

    return list(targets)


def human_to_int(h):
    '''
    converts human-readable number to integer
    e.g. 1K --> 1000
    '''

    if type(h) == int:
        return h

    units = {'': 1, 'K': 1024, 'M': 1024**2, 'G': 1024**3, 'T': 1024**4}

    try:
        h = h.upper().strip()
        i = float(''.join(c for c in h if c in string.digits + '.'))
        unit = ''.join([c for c in h if c in units.keys()])
    except (ValueError, KeyError):
        raise ValueError(f'Invalid filesize "{h}"')

    return int(i * units[unit])


def bytes_to_human(_bytes):
    '''
    converts bytes to human-readable filesize
    e.g. 1024 --> 1KB
    '''

    sizes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB']
    units = {}
    count = 0
    for size in sizes:
        units[size] = pow(1024, count)
        count +=1

    for size in sizes:
        if abs(_bytes) < 1024.0:
            if size == sizes[0]:
                _bytes = str(int(_bytes))
            else:
                _bytes = '{:.2f}'.format(_bytes)
            return '{}{}'.format(_bytes, size)
        _bytes /= 1024

    raise ValueError


def better_decode(b):

    # detect encoding with libmagic
    m = magic.Magic(mime_encoding=True)
    encoding = m.from_buffer(b)

    try:
        return b.decode(encoding)
    except Exception:
        return str(b)[2:-1]


def random_string(length):

    return ''.join(random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits) for i in range(length))


def list_files(path):

    path = Path(path)

    if path.is_file() and not path.is_symlink():
        yield path

    elif path.is_dir():
        for dir_name, dirnames, filenames in os.walk(path):
            for file in filenames:
                file = Path(dir_name) / file
                if file.is_file() and not file.is_symlink():
                    yield file


def rmdir(directory):
    '''
    Recursively remove directory
    '''
    directory = Path(directory)
    for item in directory.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()
    directory.rmdir()


import io
from .util import *
from .errors import *
from pathlib import Path


class RemoteFile():
    '''
    Represents a file on an SMB share
    Passed from a spiderling up to its parent spide
r    '''

    def __init__(self, name, share, target, size=0):

        self.share = share
        self.target = target
        self.name = name
        self.size = size
        self.smb_client = None

        file_suffix = Path(name).suffix.lower()
        self.tmp_filename = Path('/tmp/.manspider') / (random_string(15) + file_suffix)


    def get(self, smb_client=None):
        '''
        Downloads file to self.tmp_filename

        NOTE: SMBConnection() can't be passed through a multiprocessing queue
              This means that smb_client must be set after the file arrives at Spider()
        '''

        if smb_client is None and self.smb_client is None:
            raise FileRetrievalError('Please specify smb_client')

        #memfile = io.BytesIO()
        with open(str(self.tmp_filename), 'wb') as f:

            try:
                smb_client.conn.getFile(self.share, self.name, f.write)
            except Exception as e:
                smb_client.handle_impacket_error(e, self.share, self.name)
                raise FileRetrievalError(f'Error retrieving file "{str(self)}": {str(e)[:150]}')

        # reset cursor back to zero so .read() will return the whole file
        #memfile.seek(0)


    def __str__(self):

        return f'{self.target}\\{self.share}\\{self.name}'



import ntpath
import logging
from .errors import *
from contextlib import suppress
from impacket.nmb import NetBIOSError, NetBIOSTimeout
from impacket.smbconnection import SessionError, SMBConnection

# set up logging
log = logging.getLogger('manspider.smb')


class SMBClient:
    '''
    Wrapper around impacket's SMBConnection() object
    '''

    def __init__(self, server, username, password, domain, nthash, use_kerberos=False, aes_key="", dc_ip=None):

        self.server = server

        self.conn = None

        self.username = username
        self.password = password
        self.domain = domain
        self.nthash = nthash
        self.use_kerberos = use_kerberos
        self.aes_key = aes_key
        self.dc_ip = dc_ip
        self.hostname = None
        self.dns_domain = None
        if self.nthash:
            self.lmhash = 'aad3b435b51404eeaad3b435b51404ee'
        else:
            self.lmhash = ''
        self._shares = None


    def list_shares(self):
        '''
        List shares on the SMB server
        '''
        resp = self.conn.listShares()
        for i in range(len(resp)):
            sharename = resp[i]['shi1_netname'][:-1]
            log.debug(f'{self.server}: Found share: {sharename}')
            yield sharename


    @property
    def shares(self):
        if self._shares is None:
            try:
                self._shares = list(self.list_shares())
            except Exception as e:
                e = self.handle_impacket_error(e)
                log.debug(f'{self.server}: Error listing shares: {e}, retrying...')
                self.rebuild(e)
                try:
                    self._shares = list(self.list_shares())
                except Exception as e:
                    e = self.handle_impacket_error(e)
                    log.warning(f'{self.server}: Error listing shares: {e}')
                    self.rebuild(e)
        return self._shares or []


    def get_hostname(self):
        '''
        Get the hostname from the SMB connection
        '''
        try:
            conn = SMBConnection(
                self.server,
                self.server,
                None,
                445,
                timeout=10,
            )
            with suppress(Exception):
                conn.login("", "")

            if self.hostname is None:
                try:
                    # Get the server name from SMB
                    self.hostname = str(conn.getServerName()).strip().replace("\x00", "").lower()
                    if self.hostname:
                        log.debug(f'{self.server}: Got hostname: {self.hostname}')
                    else:
                        log.debug(f'{self.server}: No hostname found')
                except Exception as e:
                    log.debug(f'{self.server}: Error getting hostname from SMB: {e}')
                    self.hostname = ""

            if self.dns_domain is None:
                try:
                    self.dns_domain = str(conn.getServerDNSDomainName()).strip().replace("\x00", "").lower()
                    if self.dns_domain:
                        log.debug(f'{self.server}: Got DNS domain: {self.dns_domain}')
                    else:
                        log.debug(f'{self.server}: No DNS domain found')
                except Exception as e:
                    log.debug(f'{self.server}: Error getting DNS domain: {e}')
                    self.dns_domain = (self.domain if self.domain else "")

        except Exception as e:
            log.debug(f'{self.server}: Error getting hostname: {e}')

        return self.hostname, self.domain

    def login(self, refresh=False, first_try=True):
        '''
        Create a new SMBConnection object (if there isn't one already or if refresh is True)
        Attempt to log in, and switch to null session if logon fails
        Return True if logon succeeded
        Return False if logon failed
        '''

        target_server = self.server
        if self.use_kerberos:
            hostname, domain = self.get_hostname()
            if hostname:
                target_server = hostname
                if domain:
                    target_server = f"{hostname}.{domain}"

        if self.conn is None or refresh:
            try:
                self.conn = SMBConnection(target_server, target_server, sess_port=445, timeout=20)
            except Exception as e:
                log.debug(impacket_error(e))
                return None

            try:

                if self.username in [None, '', 'Guest'] and first_try:
                    # skip to guest / null session
                    assert False

                user_str = self.username
                if self.domain:
                    user_str = f'{self.domain}\\{self.username}'
                log.debug(f'{target_server} ({self.server}): Authenticating as "{user_str}"')

                if self.use_kerberos:
                    self.conn.kerberosLogin(
                        self.username,
                        self.password,
                        self.domain,
                        self.lmhash,
                        self.nthash,
                        self.aes_key,
                        kdcHost=self.dc_ip,
                    )
                # pass the hash if requested
                elif self.nthash and not self.password:
                    self.conn.login(
                        self.username,
                        '',
                        lmhash=self.lmhash,
                        nthash=self.nthash,
                        domain=self.domain,
                    )
                # otherwise, normal login
                else:
                    self.conn.login(
                        self.username,
                        self.password,
                        domain=self.domain,
                    )

                log.info(f'{self.server}: Successful login as "{self.username}"')
                return True

            except Exception as e:

                if type(e) != AssertionError:
                    e = self.handle_impacket_error(e, display=True)

                # try guest account, then null session if logon failed
                if first_try:

                    bad_statuses = ['LOGON_FAIL', 'PASSWORD_EXPIRED', 'LOCKED_OUT', 'SESSION_DELETED']
                    if any([s in str(e) for s in bad_statuses]):
                        for s in bad_statuses:
                            if s in str(e):
                                log.warning(f'{self.server}: {s}: {self.username}')

                    log.debug(f'{self.server}: Trying guest session')
                    self.username = 'Guest'
                    self.password = ''
                    self.domain = ''
                    self.nthash = ''
                    guest_success = self.login(refresh=True, first_try=False)
                    if not guest_success:
                        log.debug(f'{self.server}: Switching to null session')
                        self.username = ''
                        self.login(refresh=True, first_try=False)

            return False

        else:
            return True


    def ls(self, share, path):
        '''
        List files in share/path
        Raise FileListError if there's a problem
        @byt3bl33d3r it's really not that bad
        '''

        nt_path = ntpath.normpath(f'{path}\\*')

        # for every file/dir in "path"
        try:
            for f in self.conn.listPath(share, nt_path):
                # exclude current and parent directory
                if f.get_longname() not in ['', '.', '..']:
                    yield f
        except Exception as e:
            e = self.handle_impacket_error(e)
            raise FileListError(f'{e.args}: Error listing files at "{share}{nt_path}"')


    def rebuild(self, error=''):
        '''
        Rebuild our SMBConnection() if it gets borked
        '''
        log.debug(f'Rebuilding connection to {self.server} after error: {error}')
        self.login(refresh=True)


    def handle_impacket_error(self, e, share='', filename='', display=False):
        '''
        Handle arbitrary Impacket errors
        this is needed because the library doesn't implement proper inheritance for its exceptions
        '''
        resource_str = '/'.join([self.server, share, filename]).rstrip('/')

        if type(e) == KeyboardInterrupt:
            raise
        elif type(e) in (NetBIOSError, NetBIOSTimeout, BrokenPipeError, SessionError, CSessionError):
            # the connection may need to be rebuilt
            if type(e) in (SessionError, CSessionError):
                if any([x in str(e) for x in ('PASSWORD_EXPIRED',)]):
                    self.rebuild(e)
            else:
                self.rebuild(e)
        if type(e) in native_impacket_errors:
            e = impacket_error(e)
        if display:
            log.debug(f'{resource_str}: {str(e)[:150]}')

        return e



import re
import magic
import logging
from time import sleep
import subprocess as sp
from extractous import Extractor

from man_spider.lib.util import *
from man_spider.lib.logger import *

log = logging.getLogger('manspider.parser')


class FileParser:

    # don't parse files with these magic types
    magic_blacklist = [
        # PNG, JPEG, etc.
        # 'image data',
        # ZIP, GZ, etc.
        'archive data',
        # encrypted data
        'encrypted'
    ]


    def __init__(self, filters, quiet=False):
        self.init_content_filters(filters)
        self.extractor = Extractor()
        self.quiet = quiet


    def init_content_filters(self, file_content):
        '''
        Get ready to search by file content
        '''

        # strings to look for in file content
        # if empty, content is ignored
        self.content_filters = []
        for f in file_content:
            try:
                self.content_filters.append(re.compile(f, re.I))
            except re.error as e:
                log.error(f'Unsupported file content regex "{f}": {e}')
                sleep(1)
        if self.content_filters:
            content_filter_str = '"' + '", "'.join([f.pattern for f in self.content_filters]) + '"'
            log.info(f'Searching by file content: {content_filter_str}')



    def match(self, file_content):
        '''
        Finds all regex matches in file content
        '''

        for _filter in self.content_filters:
            for match in _filter.finditer(file_content):
                # ( filter, (match_start_index, match_end_index) )
                yield (_filter, match.span())


    def match_magic(self, file):
        '''
        Returns True if the file isn't of a blacklisted file type
        '''

        # get magic type
        magic_type = magic.from_file(str(file)).lower()
        for keyword in self.magic_blacklist:
            if keyword.lower() in magic_type:
                log.debug(f'Not parsing {file}: blacklisted magic type: "{keyword}"')
                return False

        return True


    def grep(self, content, pattern):

        if not self.quiet:
            try:
                '''
                GREP(1)
                    -E, --extended-regexp
                        Interpret PATTERN as an extended regular expression
                    -i, --ignore-case
                        Ignore case distinctions
                    -a, --text
                        Process a binary file as if it were text
                    -m NUM, --max-count=NUM
                        Stop  reading  a file after NUM matching lines
                '''
                grep_process = sp.Popen(
                    ['grep', '-Eiam', '5', '--color=always', pattern],
                    stdin=sp.PIPE,
                    stdout=sp.PIPE
                )
                grep_output = grep_process.communicate(content)[0]
                for line in grep_output.splitlines():
                    log.info(better_decode(line[:500]))
            except (sp.SubprocessError, OSError, IndexError):
                pass


    def parse_file(self, file, pretty_filename=None):
        '''
        Parse a file on the local filesystem
        '''

        if pretty_filename is None:
            pretty_filename = str(file)

        log.debug(f'Parsing file: {pretty_filename}')

        matches = dict()

        try:

            matches = self.textract(file, pretty_filename=pretty_filename)

        except Exception as e:
            #except (BadZipFile, textract.exceptions.CommandLineError) as e:
            if log.level <= logging.DEBUG:
                log.warning(f'Error extracting text from {pretty_filename}: {e}')
            else:
                log.warning(f'Error extracting text from {pretty_filename} (-v to debug)')
            
        return matches


    def textract(self, file, pretty_filename):
        '''
        Extracts text from a file
        Uses the textract library on specific extensions
        '''

        matches = dict()

        suffix = Path(str(file)).suffix.lower()

        # blacklist certain mime types
        if not self.match_magic(file):
            return matches

        text_content, metadata = self.extractor.extract_file_to_string(str(file))

        # try to convert to UTF-8 for grep-friendliness
        try:
            binary_content = text_content.encode('utf-8', errors='ignore')
        except Exception:
            pass

        # count the matches
        for _filter, match in self.match(text_content):
            try:
                matches[_filter] += 1
            except KeyError:
                matches[_filter] = 1

        for _filter, match_count in matches.items():
            log.info(ColoredFormatter.green(f'{pretty_filename}: matched "{_filter.pattern}" {match_count:,} times'))
            # run grep for pretty output
            if not self.quiet:
                self.grep(binary_content, _filter.pattern)

        return matches



#!/usr/bin/env python3

import sys
import pathlib
import logging
import argparse
import traceback
from .lib import *
from time import sleep
import multiprocessing


# set up logging
log = logging.getLogger('manspider')
log.setLevel(logging.INFO)


def go(options):

    log.info('MANSPIDER command executed: ' + ' '.join(sys.argv))

    try:

        # warn if --or-logic is enabled
        if options.or_logic and options.content and not all([type(t) == pathlib.PosixPath for t in options.targets]):
            log.warning('WARNING: "--or-logic" causes files to be content-searched even if filename/extension filters do not match!!')
            sleep(2)

        # exit if no filters were specified
        if not (options.filenames or options.extensions or options.exclude_extensions or options.content):
            log.error('Please specify at least one of --filenames, --content, --extensions, or --exclude-extensions')
            return

        # exit if --maxdepth is invalid
        if options.maxdepth <= 0:
            log.error('--maxdepth must be greater than zero')
            return

        log.info(f'Skipping files larger than {bytes_to_human(options.max_filesize)}')
        log.info(f'Using {options.threads:,} threads')

        manspider = MANSPIDER(options)
        manspider.start()

    except KeyboardInterrupt:
        log.critical('Interrupted')

    except Exception as e:
        if log.level <= logging.DEBUG:
            log.critical(traceback.format_exc())
        else:
            log.critical(f'Critical error (-v to debug): {e}')

    finally:
        # make sure temp files are cleaned up before exiting
        #rmdir(manspider.tmp_dir)
        pass


def main():

    interrupted = False

    examples = '''

    # EXAMPLES

    Example 1: Search the network for filenames that may contain creds
    $ manspider 192.168.0.0/24 -f passw user admin account network login logon cred -d evilcorp -u bob -p Passw0rd

    Example 2: Search for XLSX files containing "password"
    $ manspider share.evilcorp.local -c password -e xlsx -d evilcorp -u bob -p Passw0rd

    Example 3: Search for interesting file extensions
    $ manspider share.evilcorp.local -e bat com vbs ps1 psd1 psm1 pem key rsa pub reg txt cfg conf config -d evilcorp -u bob -p Passw0rd

    Example 4: Search for finance-related files
    $ manspider share.evilcorp.local --dirnames bank financ payable payment reconcil remit voucher vendor eft swift -f '[0-9]{5,}' -d evilcorp -u bob -p Passw0rd
    '''

    parser = argparse.ArgumentParser(description='Scan for juicy data on SMB shares. Matching files and logs are stored in $HOME/.manspider. All filters are case-insensitive.')
    parser.add_argument('targets', nargs='+',   type=make_targets,          help='IPs, Hostnames, CIDR ranges, or files containing targets to spider (NOTE: local searching also supported, specify directory name or keyword "loot" to search downloaded files)')
    parser.add_argument('-u', '--username',     default='',                 help='username for authentication')
    parser.add_argument('-p', '--password',     default='',                 help='password for authentication')
    parser.add_argument('-d', '--domain',       default='',                 help='domain for authentication')
    parser.add_argument('-l','--loot-dir',      default='',                 help='loot directory (default ~/.manspider/)')
    parser.add_argument('-m', '--maxdepth',     type=int,   default=10,     help='maximum depth to spider (default: 10)')
    parser.add_argument('-H', '--hash',         default='',                 help='NTLM hash for authentication')
    parser.add_argument('-k', '--kerberos',     action='store_true',       help='Use Kerberos authentication. Grabs credentials from ccache file (KRB5CCNAME) based on target parameters')
    parser.add_argument('-aesKey', '--aes-key', action='store', metavar='HEX', help='AES key to use for Kerberos Authentication (128 or 256 bits)')
    parser.add_argument('-dc-ip', '--dc-ip',    action='store', metavar='IP', help='IP Address of the domain controller. If omitted it will use the domain part (FQDN) specified in the target parameter')
    parser.add_argument('-t', '--threads',      type=int,   default=5,      help='concurrent threads (default: 5)')
    parser.add_argument('-f', '--filenames', nargs='+', default=[],         help=f'filter filenames using regex (space-separated)', metavar='REGEX')
    parser.add_argument('-e', '--extensions',nargs='+', default=[],         help='only show filenames with these extensions (space-separated, e.g. `docx xlsx` for only word & excel docs)', metavar='EXT')
    parser.add_argument('--exclude-extensions',nargs='+', default=[],       help='ignore files with these extensions', metavar='EXT')
    parser.add_argument('-c', '--content',   nargs='+', default=[],         help='search for file content using regex (multiple supported)', metavar='REGEX')
    parser.add_argument('--sharenames',      nargs='+', default=[],         help='only search shares with these names (multiple supported)', metavar='SHARE')
    parser.add_argument('--exclude-sharenames', nargs='*', default=['IPC$', 'C$', 'ADMIN$', 'PRINT$'],help='don\'t search shares with these names (multiple supported)', metavar='SHARE')
    parser.add_argument('--dirnames',      nargs='+', default=[],           help='only search directories containing these strings (multiple supported)', metavar='DIR')
    parser.add_argument('--exclude-dirnames', nargs='+', default=[],        help='don\'t search directories containing these strings (multiple supported)', metavar='DIR')
    parser.add_argument('-q', '--quiet',   action='store_true',             help='don\'t display matching file content')
    parser.add_argument('-n', '--no-download',   action='store_true',       help='don\'t download matching files')
    parser.add_argument('-mfail', '--max-failed-logons', type=int,          help='limit failed logons', metavar='INT')
    parser.add_argument('-o', '--or-logic', action='store_true',            help=f'use OR logic instead of AND (files are downloaded if filename OR extension OR content match)')
    parser.add_argument('-s', '--max-filesize', type=human_to_int, default=human_to_int('10M'), help=f'don\'t retrieve files over this size, e.g. "500K" or ".5M" (default: 10M)', metavar='SIZE')
    parser.add_argument('-v', '--verbose', action='store_true',             help='show debugging messages')
    

    syntax_error = False
    try:

        if len(sys.argv) == 1:
            parser.print_help()
            sys.exit(1)

        options = parser.parse_args()

        if options.verbose:
            log.setLevel('DEBUG')

        if options.kerberos and not "KRB5CCNAME" in os.environ:
            log.error("KRB5CCNAME is not set in the environment")
            sys.exit(1)

        # make sure extension formats are valid
        for i, extension in enumerate(options.extensions):
            if extension and not extension.startswith('.'):
                extension = f'.{extension}'
            options.extensions[i] = extension.lower()

        # make sure extension blacklist is valid
        for i, extension in enumerate(options.exclude_extensions):
            if not extension.startswith('.'):
                extension = f'.{extension}'
            options.exclude_extensions[i] = extension.lower()

        # lowercase share names
        options.sharenames = [s.lower() for s in options.sharenames]
        options.exclude_sharenames = [s.lower() for s in options.exclude_sharenames]

        # lowercase directory names
        options.dirnames = [s.lower() for s in options.dirnames]
        options.exclude_dirnames = [s.lower() for s in options.exclude_dirnames]

        # deduplicate targets
        targets = set()
        [[targets.add(t) for t in g] for g in options.targets]
        options.targets = list(targets)

        p = multiprocessing.Process(target=go, args=(options,), daemon=False)
        p.start()
        listener.start()

    except argparse.ArgumentError as e:
        syntax_error = True
        log.error(e)
        log.error('Check your syntax')
        sys.exit(2)

    except KeyboardInterrupt:
        log.critical('Interrupted')
        sys.exit(1)

    # pretty format all errors if we're not debugging
    except Exception as e:
        if log.level <= logging.DEBUG:
            log.critical(traceback.format_exc())
        else:
            log.critical(f'Critical error (-v to debug): {e}')

    finally:
        if '-h' in sys.argv or '--help' in sys.argv or len(sys.argv) == 1 or syntax_error:
            print(examples)
        sleep(1)
        try:
            # wait for go to finish
            p.join()
        except:
            pass
        try:
            # stop the log listener
            listener.stop()
        except:
            pass


if __name__ == '__main__':
    main()


import re
import sys
import queue
import threading
from time import sleep
import multiprocessing
from pathlib import Path
from .spiderling import *
from .parser import FileParser

# set up logging
log = logging.getLogger('manspider')


class MANSPIDER:

    def __init__(self, options):

        self.targets            = options.targets
        self.threads            = options.threads
        self.maxdepth           = options.maxdepth
        self.quiet              = options.quiet

        self.username           = options.username
        self.password           = options.password
        self.domain             = options.domain
        self.nthash             = options.hash
        self.use_kerberos       = options.kerberos
        self.aes_key            = options.aes_key
        self.dc_ip              = options.dc_ip
        self.max_failed_logons  = options.max_failed_logons
        self.max_filesize       = options.max_filesize

        self.share_whitelist    = options.sharenames
        self.share_blacklist    = options.exclude_sharenames

        self.dir_whitelist      = options.dirnames
        self.dir_blacklist      = options.exclude_dirnames

        self.no_download        = options.no_download

        # applies "or" logic instead of "and"
        # e.g. file is downloaded if filename OR extension OR content match
        self.or_logic           = options.or_logic

        self.extension_blacklist= options.exclude_extensions
        self.file_extensions    = options.extensions
        if self.file_extensions:
            extensions_str = '"' + '", "'.join(list(self.file_extensions)) + '"'
            log.info(f'Searching by file extension: {extensions_str}')

        self.init_filename_filters(options.filenames)
        self.parser = FileParser(options.content, quiet=self.quiet)

        self.failed_logons = 0

        self.spiderling_pool = [None] * self.threads
        self.spiderling_queue = multiprocessing.Manager().Queue()

        # prevents needing to continually instantiate new SMBClients
        # {target: SMBClient() ...}
        self.smb_client_cache = dict()

        # directory to store documents when searching contents
        self.tmp_dir = Path('/tmp/.manspider')
        self.tmp_dir.mkdir(exist_ok=True)

        # directory to store matching documents
        self.loot_dir = Path.home() / '.manspider' / 'loot'

        if(options.loot_dir):
            self.loot_dir=Path(options.loot_dir)
        
        self.loot_dir.mkdir(parents=True, exist_ok=True)

        if not options.no_download:
            log.info(f'Matching files will be downloaded to {self.loot_dir}')


    def start(self):

        for target in self.targets:
            try:
                while 1:
                    for i, process in enumerate(self.spiderling_pool):
                        # if there's room in the pool
                        if process is None or not process.is_alive():
                            # start spiderling
                            self.spiderling_pool[i] = multiprocessing.Process(
                                target=Spiderling, args=(target, self), daemon=False
                            )
                            self.spiderling_pool[i].start()
                            # success, break out of infinite loop
                            assert False
                        else:
                            # otherwise, clear the queue
                            self.check_spiderling_queue()

            except AssertionError:
                continue

            # save on CPU
            sleep(.1)

        while 1:
            self.check_spiderling_queue()
            dead_spiderlings = [s is None or not s.is_alive() for s in self.spiderling_pool]
            if all(dead_spiderlings):
                break

        # make sure the queue is empty
        self.check_spiderling_queue()



    def init_file_extensions(self, file_extensions):
        '''
        Get ready to search by file extension
        '''

        self.file_extensions = FileExtensions()
        if file_extensions:
            self.file_extensions.update(file_extensions)
            


    def init_filename_filters(self, filename_filters):
        '''
        Get ready to search by filename
        '''

        # strings to look for in filenames
        # if empty, all filenames are matched
        self.filename_filters = []
        for f in filename_filters:
            regex_str = str(f)
            try:
                if not any([f.startswith(x) for x in ['^', '.*']]):
                    regex_str = rf'.*{regex_str}'
                if not any([f.endswith(x) for x in ['$', '.*']]):
                    regex_str = rf'{regex_str}.*'
                self.filename_filters.append(re.compile(regex_str, re.I))
            except re.error as e:
                log.error(f'Unsupported filename regex "{f}": {e}')
                sleep(1)
        if self.filename_filters:
            filename_filter_str = '"' + '", "'.join([f.pattern for f in self.filename_filters]) + '"'
            log.info(f'Searching by filename: {filename_filter_str}')


    def check_spiderling_queue(self):
        '''
        Empty the spiderling queue
        '''

        while 1:
            try:
                message = self.spiderling_queue.get_nowait()
                self.process_message(message)

            except queue.Empty:
                break


    def process_message(self, message):
        '''
        Process messages from spiderlings
        Log messages, errors, files, etc.
        '''
        if message.type == 'a':
            if message.content == False:
                self.failed_logons += 1
            if self.lockout_threshold():
                log.error(f'REACHED MAXIMUM FAILED LOGONS OF {self.max_failed_logons:,}')
                log.error('KILLING EXISTING SPIDERLINGS AND CONTINUING WITH GUEST/NULL SESSIONS')
                #for spiderling in self.spiderling_pool:
                #    spiderling.kill()
                self.username = ''
                self.password = ''
                self.nthash = ''
                self.domain = ''


    def lockout_threshold(self):
        '''
        Return True if we've reached max failed logons
        '''

        if self.max_failed_logons is not None:
            if self.failed_logons >= self.max_failed_logons and self.domain:
                return True
        return False


    def get_smb_client(self, target):
        '''
        Check if we already have an smb_client cached
        If not, then create it
        '''

        smb_client = self.smb_client_cache.get(target, None)

        if smb_client is None:
            smb_client = SMBClient(
                target,
                self.username,
                self.password,
                self.domain,
                self.nthash,
                self.use_kerberos,
                self.aes_key,
                self.dc_ip
            )
            logon_result = smb_client.login()
            if logon_result == False:
                self.failed_logons += 1
            self.smb_client_cache[target] = smb_client

        return smb_client



import logging
from copy import copy
from sys import stdout
from pathlib import Path
from datetime import datetime
from multiprocessing import Queue
from logging.handlers import QueueHandler, QueueListener


### PRETTY COLORS ###


class ColoredFormatter(logging.Formatter):

    color_mapping = {
        'DEBUG':    69, # blue
        'INFO':     118, # green
        'WARNING':  208, # orange
        'ERROR':    196, # red
        'CRITICAL': 196, # red
    }

    char_mapping = {
        'DEBUG':    '*',
        'INFO':     '+',
        'WARNING':  '-',
        'ERROR':    '!',
        'CRITICAL': '!!!',
    }

    prefix = '\033[1;38;5;'
    suffix = '\033[0m'

    def __init__(self, pattern):

        super().__init__(pattern)


    def format(self, record):

        colored_record = copy(record)
        levelname = colored_record.levelname
        levelchar = self.char_mapping.get(levelname, '+')
        seq = self.color_mapping.get(levelname, 15) # default white
        colored_levelname = f'{self.prefix}{seq}m[{levelchar}]{self.suffix}'
        colored_record.levelname = colored_levelname

        return logging.Formatter.format(self, colored_record)


    @classmethod
    def green(cls, s):

        return cls.color(s)


    @classmethod
    def red(cls, s):

        return cls.color(s, level='ERROR')


    @classmethod
    def color(cls, s, level='INFO'):

        color = cls.color_mapping.get(level)
        return f'{cls.prefix}{color}m{s}{cls.suffix}'



class CustomQueueListener(QueueListener):
    '''
    Ignore errors in the monitor thread that result from a race condition when the program exits
    '''
    def _monitor(self):
        try:
            super()._monitor()
        except Exception:
            pass


### LOG TO STDERR ###

console = logging.StreamHandler(stdout)
# tell the handler to use this format
console.setFormatter(ColoredFormatter('%(levelname)s %(message)s'))

### LOG TO FILE ###

log_queue = Queue()
listener = CustomQueueListener(log_queue, console)
sender = QueueHandler(log_queue)
logging.getLogger('manspider').handlers = [sender]

logdir = Path.home() / '.manspider' / 'logs'
logdir.mkdir(parents=True, exist_ok=True)
logfile = f'manspider_{datetime.now().strftime("%m-%d-%Y")}.log'
handler = logging.FileHandler(str(logdir / logfile))
handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s'))
logging.getLogger('manspider').addHandler(handler)


import struct
import logging
from impacket.nmb import NetBIOSError, NetBIOSTimeout
from impacket.smb import SessionError, UnsupportedFeature
from impacket.smbconnection import SessionError as CSessionError

# set up logging
log = logging.getLogger('manspider')


class MANSPIDERError(Exception):
    pass

class FileRetrievalError(MANSPIDERError):
    pass

class ShareListError(MANSPIDERError):
    pass

class FileListError(MANSPIDERError):
    pass

class LogonFailure(MANSPIDERError):
    pass


native_impacket_errors = (
    struct.error,
    NetBIOSError,
    NetBIOSTimeout,
    SessionError,
    CSessionError,
    UnsupportedFeature,
)


impacket_errors = (
    OSError,
    BrokenPipeError,
) + native_impacket_errors


def impacket_error(e):
    '''
    Tries to format impacket exceptions nicely
    '''
    if type(e) in (SessionError, CSessionError):
        try:
            error_str = e.getErrorString()[0]
            e.args = (error_str,)
        except (IndexError,):
            pass
    if not e.args:
        e.args = ('',)
    return e
